;=========================================
; Program to control 1 gyroscope 
; through serial port
;-----------------------------------------
; 2002/12/20  --  lstmarcel@yahoo.fr
; http://laurent.saintmarcel.free.fr
; Target = at90s2313, 8MHz
; Compiled with avr studio from Atmel
;=========================================

.include "2313def.inc"

;------------------------------------------------------------------
;   output = PORTB (gyro order addr:[0..3])
;   input  = PORTD  pin 2=int0 = gyro data: rising edge
;		    pin 3=int1 = gyro data: falling edge
;            
;   serial = PORTD  pin 0=RX connected, 
;                   pin 1=TX not connected, 
;------------------------------------------------------------------
;
;
;------------------------------------------------------------------
;
;*****************  USER PARAMETERS *********************
;
;======================
; configure board ID
;======================
.equ	boardID		=0x00  ;0
;.equ	boardID		=0x20  ;1

;=============================
; set default servo position
;=============================
.equ	defaultServoPos =215	; 190..240
.equ	servoData0	=RAMSTART   ; take one byte by servo = 4
.equ    gyroData0       =RAMSTART+4 ; take 2 bytes by servo = 2
.equ	gyroZero0	=RAMSTART+6 ; zero take 2 bytes
.equ	gyroDiff0	=RAMSTART+8 ; diff = data-zero =>2bytes
.equ	gyroValue0	=RAMSTART+10; angle=4bytes
.equ	LONG_TIMER0	=10
.equ	pinSciAck	=6     ; pin on PORTD that is set to one 
;                                when receiving data and cleared 
;                                when SCI is ready to receive data
;------------------------------------------------------------------

;*********************  PROGRAM *************************

;=====================
; communication info
;=====================
.equ	boardBit	=0xe0 
.equ	servoSetBit	=0x10
.equ	servo0Adress	=0x08

;========================
; variables
;========================
;.equ    RAMSTART=$60    ;First On-Chip SRAM Location

;r26 & r27 reserved to XL and XH
;r28 & r29 reserved to YL and YH
;r30 & r31 reserved to ZL and ZH
.def	outputreg   = r25
.def    serialData  = r24
.def	servoMask0  = r22
.def	servoBit0   = r23
.def	addrCount0  = r0
.def	portD0	    = r2
.def	Tmp	    = r16
.def	TmpA	    = r17
.def	TmpB	    = r18
.def	Tmpl	    = r3
;===========================
; 	EEPROM
;===========================
.eseg					;EEPROM segment
.org 0

;===========================
;	SOURCE CODE
;===========================	
.cseg	;CODE segment
.org 	$000 ; reset
	rjmp __reset_
.org	INT0addr 	;/* External Interrupt0 Vector Address */
	rjmp __ext_interrupt_0_
.org	INT1addr 	;/* External Interrupt1 Vector Address */
	rjmp __ext_interrupt_1_
.org	ICP1addr 	;/* Input Capture1 Interrupt Vector Address */
	reti
.org	OC1addr 	;/* Output Compare1 Interrupt Vector Address */
	reti
.org	OVF1addr 	;/* Overflow1 Interrupt Vector Address */
	reti
.org	OVF0addr 	;/* Overflow0 Interrupt Vector Address */
	rjmp __timer0_overflow_
.org	URXCaddr 	;/* UART Receive Complete Interrupt Vector Address */
	reti
.org	UDREaddr 	;/* UART Data Register Empty Interrupt Vector Address */
	reti
.org	UTXCaddr 	;/* UART Transmit Complete Interrupt Vector Address */
	reti
.org	ACIaddr  	;/* Analog Comparator Interrupt Vector Address */
	reti
	
	
;========================
; program setup
;========================	
__reset_:
	ldi	Tmp,RAMEND	; Set initial stack ptr location at ram end
	out	SPL,Tmp
	clr	YH
	clr     ZH
	clr     XH
; define ports direction
	ldi 	Tmp, 0x0f	; PORT B = outputs b0..b3
	out 	DDRB,Tmp
	ldi 	Tmp, 0x00	; PORT C = inputs
	out 	DDRD, Tmp
; servo informations
	ldi 	servoBit0,  0x01
	ldi 	servoMask0, 0x0f ; servo enabled ; 0x00 =disable
	clr 	addrCount0
	rcall 	servoPos_init
; initialise serial port:
	rcall 	sci_init
; configure timer interrupt
	ldi 	Tmp,0x04	; CK/256
	out 	TCCR0, Tmp
	; initialise le compteur  la bonne valeur
	ldi	Tmp, LONG_TIMER0
	out 	TCNT0, Tmp
	;enable timer0 interrupt
	ldi 	Tmp,0x02         
	out 	TIMSK, Tmp
; configure timer1
	ldi 	Tmp,0x03	; CK/64 
	out 	TCCR1B, Tmp
	rcall   gyroResetAngle
; enable int0 and int1 interrupt
	ldi	Tmp, 0x0b
	out	MCUCR, Tmp
	ldi 	Tmp,0xc0         
	out 	GIMSK, Tmp
; enable interupts
	sei
; jump to real begining
	rjmp 	main
	
;========================
;  SERIAL PORT  
;========================
;**** Init serial port  ****
sci_init:	
	ldi  Tmp, 0x18	        ; RX/TX enable, 8bits, RX interruptions
	out  UCR, Tmp
	ldi  Tmp, 51	        ; 9600 baud for a 8MHz quartz
	out  UBRR, Tmp 	        ; set baud rate 
	ret
;**** send one byte on serial port : register 'serialData' ****
sci_snd:
	sbis 	USR, UDRE 		; Write Data register empty ?
	rjmp 	sci_snd 		; if No => retry
	out  	UDR, serialData	; write data on serial port
sci_snde:
	sbis 	USR, TXC 		; Transfer Data complete ?
	rjmp 	sci_snde 		; if No => retry	
	ret
;**** wait to receive one byte from serial port and store it in register 'serialData' ****
sci_rcv:
	sbis 	USR, RXC 		; Data received
	rjmp 	sci_rcv		; if No => retry
	in   	serialData, UDR	; get data
	ret
;
;========================
; servo pos
;========================
servoPos_init:
	ldi	Tmp, defaultServoPos	; default servo position time
	ldi	YL,servoData0		; move cursor to servo time
	st	Y+, Tmp
	st	Y+, Tmp
	st	Y+, Tmp
	st	Y+, Tmp
	ret
;
;=====================================
; servo control using tiner interrupt
;=====================================
;**** 4 others servo motors use timer0, and portB (b0..b3)***	
__timer0_overflow_:
	push	Tmp
	cbi	PORTB, 0
	cbi	PORTB, 1
	cbi	PORTB, 2
	cbi	PORTB, 3
endTimer0Overflow1:
	cpi	servoBit0, 0x20		; if servobit == last bit then wait a special time
	brne	endTimer0Overflow2
	ldi	Tmp, LONG_TIMER0
	out	TCNT0, Tmp
	ldi	Tmp, 0x00		; reset data
	mov	addrCount0, Tmp
	ldi	servoBit0, 0x01
	rjmp	endTimer0OverflowEND
endTimer0Overflow2:
	mov 	Tmp, servoBit0
	and	Tmp, servoMask0
	cp	Tmp, servoBit0
	brne	endTimer0Overflow3
	cpi	Tmp, 0x01
	brne	endTimer0Overflow22	
	sbi	PORTB, 0
	rjmp	endTimer0Overflow3
endTimer0Overflow22:
	cpi	Tmp, 0x02
	brne	endTimer0Overflow23	
	sbi	PORTB, 1
	rjmp	endTimer0Overflow3
endTimer0Overflow23:
	cpi	Tmp, 0x04
	brne	endTimer0Overflow24	
	sbi	PORTB, 2
	rjmp	endTimer0Overflow3
endTimer0Overflow24:
	cpi	Tmp, 0x08
	brne	endTimer0Overflow3	
	sbi	PORTB, 3
endTimer0Overflow3:
	push	YL
	ldi	YL,servoData0	;
	add	YL, addrCount0
	ld	Tmp, Y
	out	TCNT0, Tmp
	pop	YL
	; prepare cursor to next data
	inc	addrCount0
	rol	servoBit0
endTimer0OverflowEND:
	pop	Tmp
	reti
;
;=====================================
; external interrupt = gonio analyzer
;=====================================
__ext_interrupt_0_:
;	rising edge => reset timer 1
	push	Tmp
	clr	Tmp
	out	TCNT1L, Tmp
	out	TCNT1H, Tmp
	pop	Tmp
	reti
	
__ext_interrupt_1_:
;	falling edge => save timer 1 in gyroData0
;       timer1 = gyro order
	push	ZL
	push	Tmp
	ldi	ZL,gyroData0
	inc	ZL
	in      Tmp, TCNT1H
	clc	; clear carry
	ror	Tmp		; remove one bit from TCNT1 => CK/128 instead of CK/64
	st	Z, Tmp
	in      Tmp, TCNT1L
	ror	Tmp
	st	-Z, Tmp
	pop	Tmp
	pop	ZL
; compute new gyro angle
	rcall	gyroCompute	
	reti

;======================================
; Compute the current gyro angle
;======================================
gyroCompute:
	push	ZL
	push	YL
	push	XL
	push	Tmpl		
	push	Tmp		
	ldi 	ZL,gyroData0
	ldi	YL,gyroZero0
	ldi	XL,gyroDiff0
;       **** Diff=Data-Zero (on 2 bytes)****
	ld	TmpA, Z+
	ld	TmpB, Y+
	clr	Tmp
	out	SREG, Tmp
	sub	TmpA, TmpB
	st	X+,   TmpA 
	ld	TmpA, Z+
	ld	TmpB, Y+
	sbc	TmpA, TmpB
	st	X+,   TmpA 
;	**** Value=Value+Diff (on 4 bytes)****
	ldi 	ZL,gyroValue0
	ldi	YL,gyroDiff0
	ldi	XL,gyroValue0
	ld	TmpA, Z+
	ld	TmpB, Y+
	add	TmpA, TmpB
	st	X+,   TmpA 
	ld	TmpA, Z+
	ld	TmpB, Y+
	adc	TmpA, TmpB
;       Save carry
	in	Tmp,  SREG
	mov  	Tmpl, Tmp
	st	X+,   TmpA 
;       If first byte of diff is < 0 then TmpB=-1=0xff for carry
	mov	Tmp,  TmpB
	cpi	Tmp,  0x00
	brpl	gc02
	ldi	TmpB, 0xff
	rjmp	gc03
gc02:
	clr	TmpB
gc03:
	ld	TmpA, Z+	
;       Restore carry
	mov 	Tmp,  Tmpl
	out	SREG, Tmp
; 	From now, just propagate carry
	adc	TmpA, TmpB
	st	X+,   TmpA 
	ld	TmpA, Z+
	adc	TmpA, TmpB
	st	X+,   TmpA 
	pop	Tmp
	pop	Tmpl
	pop	XL
	pop	YL
	pop	ZL
	ret
	
;=============================================================
; Save current gyro order as the normal value = zero position
;=============================================================
gyroSaveZero:
	push	ZL
	push	YL
	push	Tmp
	ldi 	ZL,gyroData0
	ldi	YL,gyroZero0
	ld	Tmp, Z+
	st	Y+, Tmp	
	ld	Tmp, Z
	st	Y, Tmp	
	pop	Tmp
 	pop	YL
 	pop	ZL
	ret

;=======================================
; Clear servo angle
;=======================================	
gyroResetAngle:
	push	ZL
	push	Tmp
	ldi 	ZL,gyroValue0
 	clr	Tmp
 	st	Z+, Tmp	
 	st	Z+, Tmp	
 	st	Z+, Tmp	
 	st	Z+, Tmp
 	pop	Tmp
 	pop	ZL	
	ret

;===========================================
; send gyro angle on serial port => 4 bytes
;===========================================
gyroSendAngle0:
	ldi 	ZL,gyroValue0
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ret	
 	
;*************** debug ***************
gyroSendDiff0:
	ldi 	ZL,gyroDiff0
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ret
 	
gyroSendData0:
	ldi 	ZL,gyroData0
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ret
 	
gyroSendZero0:
	ldi 	ZL,gyroZero0
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ld	serialData, Z+
 	rcall	sci_snd	
 	ret
 	
gyroSetData0:
	ldi	ZL,gyroData0	;
	rcall	sci_rcv
	mov     Tmp, serialData
	st	Z+, Tmp
	rcall	sci_rcv
	mov     Tmp, serialData
	st	Z, Tmp
	ret

gyroDisableInterrupts:
	cli
	ret

gyroEnableInterrupts:
	sei
	ret
;
;===========================
; main program
;===========================
main:		
	rcall	sci_rcv
	cpi	serialData,0x00
	breq	comResetAll
	cpi	serialData,0x02
	breq	comSendAngle0
	cpi	serialData,0x08
	breq	comSaveZero
	cpi	serialData,0x09
	breq	comResetAngle
	cpi	serialData,0x03
	breq	comGetDiff0
	cpi	serialData,0x04
	breq	comGetData0
	cpi	serialData,0x05
	breq	comGetZero0
	cpi	serialData,0x06
	breq	comSetData0
	cpi	serialData,0x07
	breq	comCompute
	rjmp 	main
	
comSendAngle0:
	rcall 	gyroSendAngle0
	rjmp  	main
	
comSaveZero:
	rcall 	gyroSaveZero
	rjmp	main
	
comResetAngle:
	rcall 	gyroResetAngle
	rjmp	main
	
comResetAll:
	cli			; disbale interrupts
	rcall 	gyroSaveZero
	rcall 	gyroResetAngle
	sei			; enable interrupts
	rjmp	main

comGetDiff0:
	rcall 	gyroSendDiff0
	rjmp	main

comGetData0:
	rcall 	gyroSendData0
	rjmp	main

comGetZero0:
	rcall 	gyroSendZero0
	rjmp	main

comSetData0:
	rcall 	gyroSetData0
	rjmp	main
	
comCompute:	
	rcall 	gyroCompute
	rjmp	main