#include <assert.h>

#include "uart.h"
#include "Serial.h"

#define UART_DEFAULT_PORT_NAME "COM1"
static serialCL serial;

/*
 * @brief This function initialize serial port
 * @param baud rate speed for serial port,
 * @see   BAUD_RATE_9600... 
 */
void uartInit(BYTE baud)
{
	int result;
	switch(baud){
	case UART_BAUD_RATE_1200:
		result=serial.open(UART_DEFAULT_PORT_NAME, 
						1200); 
		break;
	case UART_BAUD_RATE_2400:
		result=serial.open(UART_DEFAULT_PORT_NAME, 
						2400); 
		break;
	case UART_BAUD_RATE_4800:
		result=serial.open(UART_DEFAULT_PORT_NAME, 
						4800); 
		break;
	case UART_BAUD_RATE_9600:
		result=serial.open(UART_DEFAULT_PORT_NAME, 
						9600); 
		break;
	default:
		assert(0); // invalid baud rate
	}
	if(result!=SERIAL_SUCCESS){
		MessageBox(0,"Error","Cannot open serial, it may be used by another application", MB_OK);
	}
}

/*
 * @brief This function sends one byte on serial port
 */
void uartSend(BYTE data) 
{
	BYTE buf[2];
	buf[0]=data;
	serial.write(buf, 1);
}

/*
 * @brief This function gets one byte from serial port
 * @param blocking if set to UART_DONT_WAIT_RECEIVE, if there is no byte 
 *        waiting on serial port, the function returns immediatly the value 
 *        UART_NO_RCV\n
 *        If set to UART_WAIT_RECEIVE, the function waits until a byte
 *        arrives on serial port and it returns it
 *        
 * @see UART_WAIT_RECEIVE      
 * @see UART_DONT_WAIT_RECEIVE
 * @see UART_NO_RCV
 */
BYTE uartReceive(BOOL blocking)
{
	BYTE buf[2];
	unsigned long l=1;
	if(blocking){
		serial.forceRead(buf, l);
	} else {
		buf[0]=UART_NO_RCV;
		serial.read(buf, l);
	}
	return buf[0];
}